# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests for tensorflow_graphics.datasets.features.camera_feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from tensorflow_graphics.datasets.features import camera_feature


class CameraFeatureTest(tfds.testing.FeatureExpectationsTestCase):
  """Test Cases for Camera FeatureConnector."""

  def __get_camera_params(self):
    pose = {'R': np.eye(3).astype(np.float32),
            't': np.zeros(3).astype(np.float32)}
    f = 35.
    optical_center = (640 / 2, 480 / 2)

    return pose, f, optical_center

  def test_simple_camera(self):
    """Tests camera parameters with fixed focal length, no skew and no aspect ratio."""

    expected_pose, expected_f, expected_center = self.__get_camera_params()

    expected_intrinsics = np.asarray([[expected_f, 0, expected_center[0]],
                                      [0, expected_f, expected_center[1]],
                                      [0, 0, 1]], dtype=np.float32)

    expected_camera = {'pose': expected_pose, 'intrinsics': expected_intrinsics}
    inputs = {'f': expected_f, 'optical_center': expected_center,
              'pose': expected_pose}
    lookat_inputs = {
        'f': expected_f,
        'optical_center': expected_center,
        'pose': {
            'look_at': np.array([0, 0, -1], dtype=np.float32),
            'up': np.array([0, 1, 0], dtype=np.float32),
            'position': np.array([0, 0, 0], dtype=np.float32)
        }
    }

    raising_pose_entry = {
        'f': expected_f,
        'optical_center': expected_center,
        'pose': np.eye(4)
    }

    raising_pose_inputs = {
        'f': expected_f,
        'optical_center': expected_center,
        'pose': {'rot': np.eye(3), 'trans': np.zeros(3)}
    }
    raising_lookat_inputs = {
        'f': expected_f,
        'optical_center': expected_center,
        'pose': {
            'l': np.array([0, 0, -1], dtype=np.float32),
            'up': np.array([0, 1, 0], dtype=np.float32),
            'C': np.array([0, 0, 0], dtype=np.float32)
        }
    }

    self.assertFeature(
        feature=camera_feature.Camera(),
        shape={
            'pose': {
                'R': (3, 3),
                't': (3,)
            },
            'intrinsics': (3, 3)
        },
        dtype={
            'pose': {
                'R': tf.float32,
                't': tf.float32
            },
            'intrinsics': tf.float32
        },
        tests=[
            tfds.testing.FeatureExpectationItem(
                value=inputs,
                expected=expected_camera,
            ),
            tfds.testing.FeatureExpectationItem(
                value=lookat_inputs,
                expected=expected_camera
            ),
            tfds.testing.FeatureExpectationItem(
                value=raising_pose_inputs,
                raise_cls=ValueError,
                raise_msg='Wrong keys for pose feature provided'
            ),
            tfds.testing.FeatureExpectationItem(
                value=raising_lookat_inputs,
                raise_cls=ValueError,
                raise_msg='Wrong keys for pose feature provided'
            ),
            tfds.testing.FeatureExpectationItem(
                value=raising_pose_entry,
                raise_cls=ValueError,
                raise_msg='Pose needs to be a dictionary'
            ),
        ],
    )

  def test_camera_with_aspect_ratio_and_skew(self):
    """Tests camera parameters with fixed focal length, aspect_ratio and skew."""

    expected_pose, expected_f, expected_center = self.__get_camera_params()
    expected_aspect_ratio = expected_center[0] / expected_center[1]
    expected_skew = 0.6
    expected_intrinsics = np.asarray(
        [[expected_f, expected_skew, expected_center[0]],
         [0, expected_aspect_ratio * expected_f, expected_center[1]],
         [0, 0, 1]], dtype=np.float32)

    expected_camera = {'pose': expected_pose, 'intrinsics': expected_intrinsics}
    inputs = {'f': expected_f,
              'optical_center': expected_center,
              'skew': expected_skew,
              'aspect_ratio': expected_aspect_ratio,
              'pose': expected_pose}

    self.assertFeature(
        feature=camera_feature.Camera(),
        shape={
            'pose': {
                'R': (3, 3),
                't': (3,)
            },
            'intrinsics': (3, 3)
        },
        dtype={
            'pose': {
                'R': tf.float32,
                't': tf.float32
            },
            'intrinsics': tf.float32
        },
        tests=[
            tfds.testing.FeatureExpectationItem(
                value=inputs,
                expected=expected_camera,
            ),
        ],
    )

  def test_full_camera_calibration_matrix(self):
    """Tests camera parameters with different focal length per camera axis and skew."""

    expected_pose, _, expected_optical_center = self.__get_camera_params()
    expected_skew = 0.6
    expected_f = (35., 40.)
    expected_intrinsics = np.array(
        [[expected_f[0], expected_skew, expected_optical_center[0]],
         [0, expected_f[1], expected_optical_center[1]],
         [0, 0, 1]], dtype=np.float32)

    expected_camera = {'pose': expected_pose, 'intrinsics': expected_intrinsics}
    inputs = {'f': expected_f,
              'optical_center': expected_optical_center,
              'skew': expected_skew, 'pose': expected_pose}

    raising_inputs = {'f': expected_f,
                      'aspect_ratio': 1.5,
                      'optical_center': expected_optical_center,
                      'skew': expected_skew, 'pose': expected_pose}
    self.assertFeature(
        feature=camera_feature.Camera(),
        shape={
            'pose': {
                'R': (3, 3),
                't': (3,)
            },
            'intrinsics': (3, 3)
        },
        dtype={
            'pose': {
                'R': tf.float32,
                't': tf.float32
            },
            'intrinsics': tf.float32
        },
        tests=[
            tfds.testing.FeatureExpectationItem(
                value=inputs,
                expected=expected_camera,
            ),
            tfds.testing.FeatureExpectationItem(
                value=raising_inputs,
                raise_cls=ValueError,
                raise_msg='If aspect ratio is provided, f needs to '
                          'be a single float',
            ),
        ],
    )


if __name__ == '__main__':
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for point light."""

import math

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.rendering.light import point_light
from tensorflow_graphics.util import test_case


def fake_brdf(incoming_light_direction, outgoing_light_direction,
              surface_point_normal):
  del incoming_light_direction, surface_point_normal  # Unused.
  return outgoing_light_direction


def returning_zeros_brdf(incoming_light_direction, outgoing_light_direction,
                         surface_point_normal):
  del incoming_light_direction, outgoing_light_direction  # Unused.
  return tf.zeros_like(surface_point_normal)


def random_tensor(tensor_shape):
  return np.random.uniform(low=-100.0, high=100.0, size=tensor_shape)


class PointLightTest(test_case.TestCase):

  @parameterized.parameters(
      # Light direction is parallel to the surface normal.
      ([1.], [[0., 0., 1.]], [2., 0., 0.], [1.0 / (4. * math.pi), 0., 0.]),
      # Light direction is parallel to the surface normal and the reflected
      # light fall off is included in the calculation.
      ([1.], [[0., 0., 1.]], [2., 0., 0.], \
       [0.25 / (4. * math.pi), 0., 0.], True),
      # Light direction is perpendicular to the surface normal.
      ([1.], [[3., 0., 0.]], [1., 2., 3.], [0., 0., 0.]),
      # Angle between surface normal and the incoming light direction is pi/3.
      ([1.], [[math.sqrt(3), 0., 1.]], \
       [0., 1., 0.], [0., 0.125 / (4. * math.pi), 0.]),
      # Angle between surface normal and the incoming light direction is pi/4.
      ([1.], [[0., 1., 1.]], [1., 1., 0.],
       [0.25 / (4. * math.pi), 0.25 / (4. * math.pi), 0.]),
      # Light has 3 radiances.
      ([2., 4., 1.], [[0., 1., 1.]], [1., 1., 0.],
       [0.5 / (4. * math.pi), 1. / (4. * math.pi), 0.]),
      # Light is behind the surface.
      ([1.], [[0., 0., -2.]], [7., 0., 0.], [0., 0., 0.]),
      # Observation point is behind the surface.
      ([1.], [[0., 0., 2.]], [5., 0., -2.], [0., 0., 0.]),
      # Light and observation point are behind the surface.
      ([1.], [[0., 0., -2.]], [5., 0., -2.], [0., 0., 0.]),
  )
  def test_estimate_radiance_preset(self,
                                    light_radiance,
                                    light_pos,
                                    observation_pos,
                                    expected_result,
                                    reflected_light_fall_off=False):
    """Tests the output of estimate radiance function with various parameters.

    In this test the point on the surface is always [0, 0, 0] ,the surface
    normal is [0, 0, 1] and the fake brdf function returns the (normalized)
    direction of the outgoing light as its output.

    Args:
     light_radiance: An array of size K representing the point light radiances.
     light_pos: An array of size [3,] representing the point light positions.
     observation_pos: An array of size [3,] representing the observation point.
     expected_result: An array of size [3,] representing the expected result of
       the estimated reflected radiance function.
     reflected_light_fall_off: A boolean specifying whether or not to include
       the fall off of the reflected light in the calculation. Defaults to
       False.
    """
    tensor_size = np.random.randint(1, 3) + 1
    tensor_shape = np.random.randint(1, 10, size=tensor_size).tolist()
    lights_tensor_size = np.random.randint(1, 3) + 1
    lights_tensor_shape = np.random.randint(
        1, 10, size=lights_tensor_size).tolist()
    point_light_radiance = np.tile(light_radiance, lights_tensor_shape + [1])
    point_light_position = np.tile(light_pos, lights_tensor_shape + [1])
    surface_point_normal = np.tile([0.0, 0.0, 1.0], tensor_shape + [1])
    surface_point_position = np.tile([0.0, 0.0, 0.0], tensor_shape + [1])
    observation_point = np.tile(observation_pos, tensor_shape + [1])
    expected = np.tile(expected_result,
                       tensor_shape + lights_tensor_shape + [1])

    pred = point_light.estimate_radiance(
        point_light_radiance,
        point_light_position,
        surface_point_position,
        surface_point_normal,
        observation_point,
        fake_brdf,
        reflected_light_fall_off=reflected_light_fall_off)

    self.assertAllClose(expected, pred)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_estimate_radiance_jacobian_random(self):
    """Tests the Jacobian of the point lighting equation."""
    tensor_size = np.random.randint(1, 3)
    tensor_shape = np.random.randint(1, 10, size=tensor_size).tolist()
    light_tensor_size = np.random.randint(1, 3)
    lights_tensor_shape = np.random.randint(
        1, 10, size=light_tensor_size).tolist()
    point_light_radiance_init = random_tensor(lights_tensor_shape + [1])
    point_light_position_init = random_tensor(lights_tensor_shape + [3])
    surface_point_position_init = random_tensor(tensor_shape + [3])
    surface_point_normal_init = random_tensor(tensor_shape + [3])
    observation_point_init = random_tensor(tensor_shape + [3])

    def estimate_radiance_fn(point_light_position, surface_point_position,
                             surface_point_normal, observation_point):
      return point_light.estimate_radiance(point_light_radiance_init,
                                           point_light_position,
                                           surface_point_position,
                                           surface_point_normal,
                                           observation_point, fake_brdf)

    self.assert_jacobian_is_correct_fn(estimate_radiance_fn, [
        point_light_position_init, surface_point_position_init,
        surface_point_normal_init, observation_point_init
    ])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_estimate_radiance_jacobian_preset(self):
    """Tests the Jacobian of the point lighting equation.

    Verifies that the Jacobian of the point lighting equation is correct when
    the light direction is orthogonal to the surface normal.
    """
    delta = 1e-5
    point_light_radiance_init = np.array(1.0).reshape((1, 1))
    point_light_position_init = np.array((delta, 1.0, 0.0)).reshape((1, 3))
    surface_point_position_init = np.array((0.0, 0.0, 0.0))
    surface_point_normal_init = np.array((1.0, 0.0, 0.0))
    observation_point_init = np.array((delta, 3.0, 0.0))

    def estimate_radiance_fn(point_light_position, surface_point_position,
                             surface_point_normal, observation_point):
      return point_light.estimate_radiance(point_light_radiance_init,
                                           point_light_position,
                                           surface_point_position,
                                           surface_point_normal,
                                           observation_point, fake_brdf)

    self.assert_jacobian_is_correct_fn(estimate_radiance_fn, [
        point_light_position_init, surface_point_position_init,
        surface_point_normal_init, observation_point_init
    ])

  @parameterized.parameters(
      ((1, 1), (1, 3), (3,), (3,), (3,)),
      ((4, 1, 1), (4, 1, 3), (1, 3), (1, 3), (1, 3)),
      ((3, 2, 1), (3, 2, 3), (2, 3), (2, 3), (2, 3)),
      ((1, 1), (3,), (1, 3), (1, 2, 3), (1, 3)),
      ((4, 5, 1), (3, 4, 5, 3), (1, 3), (1, 2, 2, 3), (1, 2, 3)),
      ((1,), (1, 2, 2, 3), (1, 2, 3), (1, 3), (3,)),
  )
  def test_estimate_radiance_shape_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        point_light.estimate_radiance, shape, brdf=returning_zeros_brdf)

  @parameterized.parameters(
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 1), (3,), (3,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (5, 1), (5, 2), (3,), (3,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 4), (3,), (3,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (1,), (3,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (2,), (3,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (4,), (3,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (3,), (1,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (3,), (2,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (3,), (4,),
       (3,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (3,), (3,),
       (4,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (3,), (3,),
       (2,)),
      ("must have exactly 3 dimensions in axis -1", (1, 1), (1, 3), (3,), (3,),
       (1,)),
      ("Not all batch dimensions are broadcast-compatible.", (1, 3, 1),
       (1, 3, 3), (2, 3), (4, 3), (3,)),
      ("Not all batch dimensions are broadcast-compatible.", (1, 3, 1),
       (1, 4, 3), (2, 3), (3,), (3,)),
  )
  def test_estimate_radiance_shape_exception_raised(self, error_msg, *shape):
    """Tests that the shape exception is raised."""
    self.assert_exception_is_raised(
        point_light.estimate_radiance,
        error_msg,
        shape,
        brdf=returning_zeros_brdf)

  def test_estimate_radiance_value_exceptions_raised(self):
    """Tests that the value exceptions are raised correctly."""
    point_light_radiance = random_tensor(tensor_shape=(1, 1))
    point_light_position = random_tensor(tensor_shape=(1, 3))
    surface_point_position = random_tensor(tensor_shape=(3,))
    surface_point_normal = random_tensor(tensor_shape=(3,))
    observation_point = random_tensor(tensor_shape=(3,))

    # Verify that an InvalidArgumentError is raised as the given
    # surface_point_normal is not normalized.
    with self.assertRaises(tf.errors.InvalidArgumentError):
      self.evaluate(
          point_light.estimate_radiance(point_light_radiance,
                                        point_light_position,
                                        surface_point_position,
                                        surface_point_normal, observation_point,
                                        returning_zeros_brdf))


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""This module implements axis-angle functionalities.

The axis-angle representation is defined as $$\theta\mathbf{a}$$, where
$$\mathbf{a}$$ is a unit vector indicating the direction of rotation and
$$\theta$$ is a scalar controlling the angle of rotation. It is important to
note that the axis-angle does not perform rotation by itself, but that it can be
used to rotate any given vector $$\mathbf{v} \in {\mathbb{R}^3}$$ into
a vector $$\mathbf{v}'$$ using the Rodrigues' rotation formula:

$$\mathbf{v}'=\mathbf{v}\cos(\theta)+(\mathbf{a}\times\mathbf{v})\sin(\theta)
+\mathbf{a}(\mathbf{a}\cdot\mathbf{v})(1-\cos(\theta)).$$

More details about the axis-angle formalism can be found on [this page.]
(https://en.wikipedia.org/wiki/Axis%E2%80%93angle_representation)

Note: Some of the functions defined in the module expect
a normalized axis $$\mathbf{a} = [x, y, z]^T$$ as inputs where
$$x^2 + y^2 + z^2 = 1$$.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.geometry.transformation import quaternion as quaternion_lib
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def from_euler(angles, name=None):
  r"""Converts Euler angles to an axis-angle representation.

  Note:
    The conversion is performed by first converting to a quaternion
    representation, and then by converting the quaternion to an axis-angle.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    angles: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the three Euler angles. `[A1, ..., An, 0]` is the angle about
      `x` in radians `[A1, ..., An, 1]` is the angle about `y` in radians and
      `[A1, ..., An, 2]` is the angle about `z` in radians.
    name: A name for this op that defaults to "axis_angle_from_euler".

  Returns:
    A tuple of two tensors, respectively of shape `[A1, ..., An, 3]` and
    `[A1, ..., An, 1]`, where the first tensor represents the axis, and the
    second represents the angle. The resulting axis is a normalized vector.
  """
  with tf.compat.v1.name_scope(name, "axis_angle_from_euler", [angles]):
    quaternion = quaternion_lib.from_euler(angles)
    return from_quaternion(quaternion)


def from_euler_with_small_angles_approximation(angles, name=None):
  r"""Converts small Euler angles to an axis-angle representation.

  Under the small angle assumption, $$\sin(x)$$ and $$\cos(x)$$ can be
  approximated by their second order Taylor expansions, where
  $$\sin(x) \approx x$$ and $$\cos(x) \approx 1 - \frac{x^2}{2}$$.
  In the current implementation, the smallness of the angles is not verified.

  Note:
    The conversion is performed by first converting to a quaternion
    representation, and then by converting the quaternion to an axis-angle.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    angles: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the three small Euler angles. `[A1, ..., An, 0]` is the angle
      about `x` in radians `[A1, ..., An, 1]` is the angle about `y` in radians
      and `[A1, ..., An, 2]` is the angle about `z` in radians.
    name: A name for this op that defaults to
      "axis_angle_from_euler_with_small_angles_approximation".

  Returns:
    A tuple of two tensors, respectively of shape `[A1, ..., An, 3]` and
    `[A1, ..., An, 1]`, where the first tensor represents the axis, and the
    second represents the angle. The resulting axis is a normalized vector.
  """
  with tf.compat.v1.name_scope(
      name, "axis_angle_from_euler_with_small_angles_approximation", [angles]):
    quaternion = quaternion_lib.from_euler_with_small_angles_approximation(
        angles)
    return from_quaternion(quaternion)


def from_quaternion(quaternion, name=None):
  """Converts a quaternion to an axis-angle representation.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion: A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    name: A name for this op that defaults to "axis_angle_from_quaternion".

  Returns:
    Tuple of two tensors of shape `[A1, ..., An, 3]` and `[A1, ..., An, 1]`,
    where the first tensor represents the axis, and the second represents the
    angle. The resulting axis is a normalized vector.

  Raises:
    ValueError: If the shape of `quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "axis_angle_from_quaternion",
                               [quaternion]):
    quaternion = tf.convert_to_tensor(value=quaternion)

    shape.check_static(
        tensor=quaternion, tensor_name="quaternion", has_dim_equals=(-1, 4))
    quaternion = asserts.assert_normalized(quaternion)

    # This prevents zero norm xyz and zero w, and is differentiable.
    quaternion += asserts.select_eps_for_addition(quaternion.dtype)
    xyz, w = tf.split(quaternion, (3, 1), axis=-1)
    norm = tf.norm(tensor=xyz, axis=-1, keepdims=True)
    angle = 2.0 * tf.atan2(norm, tf.abs(w))
    axis = safe_ops.safe_unsigned_div(safe_ops.nonzero_sign(w) * xyz, norm)
    return axis, angle


def from_rotation_matrix(rotation_matrix, name=None):
  """Converts a rotation matrix to an axis-angle representation.

  Note:
    In the current version the returned axis-angle representation is not unique
    for a given rotation matrix. Since a direct conversion would not really be
    faster, we first transform the rotation matrix to a quaternion, and finally
    perform the conversion from that quaternion to the corresponding axis-angle
    representation.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    rotation_matrix: A tensor of shape `[A1, ..., An, 3, 3]`, where the last two
      dimensions represent a rotation matrix.
    name: A name for this op that defaults to "axis_angle_from_rotation_matrix".

  Returns:
    A tuple of two tensors, respectively of shape `[A1, ..., An, 3]` and
    `[A1, ..., An, 1]`, where the first tensor represents the axis, and the
    second represents the angle. The resulting axis is a normalized vector.

  Raises:
    ValueError: If the shape of `rotation_matrix` is not supported.
  """
  with tf.compat.v1.name_scope(name, "axis_angle_from_rotation_matrix",
                               [rotation_matrix]):
    rotation_matrix = tf.convert_to_tensor(value=rotation_matrix)

    shape.check_static(
        tensor=rotation_matrix,
        tensor_name="rotation_matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-2, 3), (-1, 3)))
    rotation_matrix = rotation_matrix_3d.assert_rotation_matrix_normalized(
        rotation_matrix)

    quaternion = quaternion_lib.from_rotation_matrix(rotation_matrix)
    return from_quaternion(quaternion)


def inverse(axis, angle, name=None):
  """Computes the axis-angle that is the inverse of the input axis-angle.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    axis: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a normalized axis.
    angle: A tensor of shape `[A1, ..., An, 1]` where the last dimension
      represents an angle.
    name: A name for this op that defaults to "axis_angle_inverse".

  Returns:
    A tuple of two tensors, respectively of shape `[A1, ..., An, 3]` and
    `[A1, ..., An, 1]`, where the first tensor represents the axis, and the
    second represents the angle. The resulting axis is a normalized vector.

  Raises:
    ValueError: If the shape of `axis` or `angle` is not supported.
  """
  with tf.compat.v1.name_scope(name, "axis_angle_inverse", [axis, angle]):
    axis = tf.convert_to_tensor(value=axis)
    angle = tf.convert_to_tensor(value=angle)

    shape.check_static(tensor=axis, tensor_name="axis", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=angle, tensor_name="angle", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(axis, angle),
        tensor_names=("axis", "angle"),
        last_axes=-2,
        broadcast_compatible=True)

    axis = asserts.assert_normalized(axis)
    return axis, -angle


def is_normalized(axis, angle, atol=1e-3, name=None):
  """Determines if the axis-angle is normalized or not.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    axis: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a normalized axis.
    angle: A tensor of shape `[A1, ..., An, 1]` where the last dimension
      represents an angle.
    atol: The absolute tolerance parameter.
    name: A name for this op that defaults to "axis_angle_is_normalized".

  Returns:
    A tensor of shape `[A1, ..., An, 1]`, where False indicates that the axis is
    not normalized.
  """
  with tf.compat.v1.name_scope(name, "axis_angle_is_normalized", [axis, angle]):
    axis = tf.convert_to_tensor(value=axis)
    angle = tf.convert_to_tensor(value=angle)

    shape.check_static(tensor=axis, tensor_name="axis", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=angle, tensor_name="angle", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(axis, angle),
        tensor_names=("axis", "angle"),
        last_axes=-2,
        broadcast_compatible=True)

    norms = tf.norm(tensor=axis, axis=-1, keepdims=True)
    return tf.abs(norms - 1.) < atol


def rotate(point, axis, angle, name=None):
  r"""Rotates a 3d point using an axis-angle by applying the Rodrigues' formula.

  Rotates a vector $$\mathbf{v} \in {\mathbb{R}^3}$$ into a vector
  $$\mathbf{v}' \in {\mathbb{R}^3}$$ using the Rodrigues' rotation formula:

  $$\mathbf{v}'=\mathbf{v}\cos(\theta)+(\mathbf{a}\times\mathbf{v})\sin(\theta)
  +\mathbf{a}(\mathbf{a}\cdot\mathbf{v})(1-\cos(\theta)).$$

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    point: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a 3d point to rotate.
    axis: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a normalized axis.
    angle: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents an angle.
    name: A name for this op that defaults to "axis_angle_rotate".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents
    a 3d point.

  Raises:
    ValueError: If `point`, `axis`, or `angle` are of different shape or if
    their respective shape is not supported.
  """
  with tf.compat.v1.name_scope(name, "axis_angle_rotate", [point, axis, angle]):
    point = tf.convert_to_tensor(value=point)
    axis = tf.convert_to_tensor(value=axis)
    angle = tf.convert_to_tensor(value=angle)

    shape.check_static(
        tensor=point, tensor_name="point", has_dim_equals=(-1, 3))
    shape.check_static(tensor=axis, tensor_name="axis", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=angle, tensor_name="angle", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(point, axis, angle),
        tensor_names=("point", "axis", "angle"),
        last_axes=-2,
        broadcast_compatible=True)
    axis = asserts.assert_normalized(axis)

    cos_angle = tf.cos(angle)
    axis_dot_point = vector.dot(axis, point)
    return point * cos_angle + vector.cross(
        axis, point) * tf.sin(angle) + axis * axis_dot_point * (1.0 - cos_angle)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()
